local PlayerClass = select(2, UnitClass("player"))

local Talented = LibStub("AceAddon-3.0"):NewAddon("Talented",
	"AceConsole-3.0", "AceComm-3.0", "AceHook-3.0", "AceEvent-3.0", "AceSerializer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Talented")

Talented.max_talent_points = 71

Talented.prev_Print = Talented.Print
function Talented:Print(s, ...)
	if s:find("%", nil, true) then
		self:prev_Print(s:format(...))
	else
		self:prev_Print(s, ...)
	end
end

function Talented:Debug(...)
	if self.db.profile.debug then
		self:Print(...)
	end
end

function Talented:MakeTarget()
	local name = self.db.char.target
	local src = name and self.db.global.templates[name]
	if not src then
		if name then
			self.db.char.target = nil
		end
		return
	end

	local target = self.target
	if not target then
		target = {}
		self.target = target
	end
	self:CopyPackedTemplate(src, target)

	if not self:ValidateTemplate(target) or target.class ~= PlayerClass then
		self.target = nil
		self.db.char.target = nil
	end
end

function Talented:SetTemplateMode(mode, template)
	if mode then
		if not template then
			local p = self.db.global
			local t = self.db.profile.last_template
			if t then
				template = p.templates[t]
			end
			if not template then
				local _
				_, template = next(p.templates)
			end
		end
		self:SetTemplate(template or self:CreateEmptyTemplate())
	else
		self:SetTemplate()
	end
	local b = self.base
	if b then
		local cb = b.checkbox
		if mode then
			b.editname:Show()
			cb.label:SetText(L["Edit template"])
			cb.tooltip = L["Toggle editing of talents."]
		else
			b.editname:Hide()
			cb.label:SetText(L["Edit talents"])
			cb.tooltip = L["Toggle editing of talents."]
		end
		cb:SetChecked(self.mode == "edit")
	end
end

function Talented:IsTemplateMode()
	return self.template ~= self.current
end

function Talented:OnDisable()
	self:SetTemplateMode()
	self:UnhookInspectUI()
end

function Talented:GetMode()
	return self.mode
end

function Talented:ApplyCurrentTemplate()
	local template = self.template
	if template.class ~= PlayerClass then
		return
	end
	local count = 0
	-- check if enough talent points are available
	local available = UnitCharacterPoints("player")
	for tab, tree in ipairs(self:GetTalentInfo(PlayerClass)) do
		for index, info in ipairs(tree.talents) do
			local delta = template[tab][index] - self.current[tab][index]
			if delta > 0 then
				count = count + delta
			end
		end
	end
	if count > available then
		self:Print(L["Sorry, I can't apply this template because you don't have enough talent points available (need %d)!"], count)
		self.mode = "view"
		self:UpdateView()
	else
		-- let's go !
		self:Print(L["Please wait while I set your talents..."])
		self:EnableUI(false)
		self:CHARACTER_POINTS_CHANGED()
	end
end

function Talented:SetMode(mode)
	if self.template == self.current and mode ~= "edit" then
		mode = "view"
	end
	if self.mode ~= mode then
		self.mode = mode
		if mode == "apply" then
			self:ApplyCurrentTemplate()
		elseif mode == "delete" then
			self:DeleteCurrentTemplate()
		end
		self:UpdateView()
	end
	local cb = self.base and self.base.checkbox
	if cb then
		cb:SetChecked(mode == "edit")
	end
end

function Talented:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("TalentedDB", self.defaults)

	local src, dst = self.db.profile.templates, self.db.global.templates
	if src then
		local import = (" from (%s)"):format(UnitName"player")
		for k, v in pairs(src) do
			k = k..import
			if not dst[k] then
				dst[k] = v
				v.name = k
			end
		end
		self.db.profile.templates = nil
		self:Print("Successfully switched to Global template, enjoy sharing your templates between characters!")
	end
	for name, template in pairs(dst) do
		template.name = name
	end

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		self.options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
		self.options.args.profiles.order = 200
	end

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Talented", self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Talented", "Talented")
	self:RegisterChatCommand("talented", "OnChatCommand")

	self:RegisterComm("Talented")
end

function Talented:OnChatCommand(input)
	if not input or input:trim() == "" then
		self:OpenOptionsFrame()
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(self, "talented", "Talented", input)
	end
end

function Talented:DeleteCurrentTemplate()
	local template = self.template
	if template == self.current then return end
	local templates = self.db.global.templates
	templates[template.name] = nil
	template = next(templates)
	self:SetTemplateMode(template ~= nil, template)
end

function Talented:UpdateTemplateName(template, newname)
	if self.db.global.templates[newname] or
			self.mode == "apply" or
			self.current == template then
		return
	end

	local oldname = template.name
	template.name = newname
	local t = self.db.global.templates
	t[newname] = template
	t[oldname] = nil
end

do
	local function new(templates, name, class)
		local count = 0
		local template = {
			name = name,
			class = class
		}
		while templates[template.name] do
			count = count + 1
			template.name = format(L["%s (%d)"], name, count)
		end
		templates[template.name] = template
		return template
	end

	local function copy(dst, src)
		dst.class = src.class
		if src.code then
			dst.code = src.code
			return
		else
			for tab, tree in ipairs(Talented:GetTalentInfo(src.class)) do
				local s, d = src[tab], {}
				dst[tab] = d
				for index in ipairs(tree.talents) do
					d[index] = s[index]
				end
			end
		end
	end

	function Talented:ImportFromOther(name, src)
		if not self:GetTalentInfo(src.class) then return end

		local dst = new(self.db.global.templates, name, src.class)

		copy(dst, src)

		self:OpenTemplate(dst)

		return dst
	end

	function Talented:CopyTemplate(src)
		local dst = new(self.db.global.templates, format(L["Copy of %s"], src.name), src.class)

		copy(dst, src)

		return dst
	end

	function Talented:CreateEmptyTemplate(class)
		if not class then class = PlayerClass end
		local template = new(self.db.global.templates, L["Empty"], class)

		local info = self:GetTalentInfo(class)
		if not info then return end

		for tab, tree in ipairs(info) do
			local t = {}
			template[tab] = t
			for index in ipairs(tree.talents) do
				t[index] = 0
			end
		end

		return template
	end

	Talented.importers = {}
	Talented.exporters = {}
	function Talented:ImportTemplate(url)
		local dst, result = new(self.db.global.templates, L["Imported"])
		for pattern, method in pairs(self.importers) do
			if url:find(pattern) then
				result = method(self, url, dst)
				if result then break end
			end
		end
		if result then
			if not self:ValidateTemplate(dst) then
				self:Print(L["The given template is not a valid one!"])
			else
				return dst
			end
		else
			self:Print(L["\"%s\" does not appear to be a valid URL!"], url)
		end
	end
end

function Talented:OpenTemplate(template)
	local base = self:CreateBaseFrame()
	if not self.current then
		self:UpdateCurrentTemplate()
	end
	self:SetTemplateMode(true, template)
	if not base:IsVisible() then
		ShowUIPanel(base)
	end
end

function Talented:SetTemplate(template)
	if not template then template = self.current end
	local old = self.template
	if template ~= old then
		self:UnpackTemplate(template)
		if not self:CreateTalentMatrix(template.class) then
			assert(template ~= self.current)
			return self:SetTemplate() -- default to current
		end
		self:PackTemplate(old)
		self.template = template
	end
	if template and template ~= self.current then
		self.db.profile.last_template = template.name
	end
	self:SetMode(self:GetDefaultMode())
	self:UpdateView()
end

function Talented:GetDefaultMode()
	return self.db.profile.always_edit and "edit" or "view"
end

function Talented:OnEnable()
	self:RawHook("ToggleTalentFrame", true)
	self:SecureHook("UpdateTalentButton")
	self:RegisterEvent("PLAYER_LOGOUT")
	self:CheckHookInspectUI()
end

function Talented:PLAYER_LOGOUT()
	self:SetTemplate()
end

function Talented:CHARACTER_POINTS_CHANGED()
	self:UpdateCurrentTemplate()
	self:UpdateView()
	if self.mode == "apply" then
		self:ApplyNextTalentPoint()
	end
end

function Talented:UpdateTalentButton()
	if self.db.profile.donthide and UnitLevel("player") < 10 then
		TalentMicroButton:Show()
		QuestLogMicroButton:SetPoint("BOTTOMLEFT", "TalentMicroButton", "BOTTOMRIGHT", -2, 0)
	end
end

function Talented:ToggleTalentFrame()
	local frame = self.base
	if not frame or not frame:IsVisible() then
		self:Update()
		ShowUIPanel(self.base)
	else
		HideUIPanel(frame)
	end
end

function Talented:Update()
	self:CreateBaseFrame()
	self:UpdateCurrentTemplate()
	if not self.template then
		self:SetTemplate(self.current)
	end
	self:UpdateView()
end

_G.Talented = Talented
